"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CodeMirror = require("codemirror");
const lsp_editor_adapter_1 = require("lsp-editor-adapter");
const utils_1 = require("../utils");
const converter_1 = require("../converter");
const lsp_1 = require("../lsp");
const positioning_1 = require("../positioning");
// TODO: settings
const hover_modifier = 'Control';
const default_severity = 2;
class CodeMirrorAdapterExtension extends lsp_editor_adapter_1.CodeMirrorAdapter {
    constructor(connection, options, editor, create_tooltip, invoke_completer, virtual_document) {
        super(connection, options, editor);
        this.virtual_document = virtual_document;
        this.marked_diagnostics = new Map();
        this.create_tooltip = create_tooltip;
        this.invoke_completer = invoke_completer;
        this.unique_editor_ids = new utils_1.DefaultMap(() => this.unique_editor_ids.size);
        // @ts-ignore
        let listeners = this.editorListeners;
        let wrapper = this.editor.getWrapperElement();
        this.editor.addEventListener('mouseleave', 
        // TODO: remove_tooltip() but allow the mouse to leave if it enters the tooltip
        //  (a bit tricky: normally we would just place the tooltip within, but it was designed to be attached to body)
        this.remove_range_highlight.bind(this));
        wrapper.addEventListener('mouseleave', this.remove_range_highlight.bind(this));
        // detach the adapters contextmenu
        wrapper.removeEventListener('contextmenu', listeners.contextmenu);
        // TODO: actually we only need the connection...
        //  the tooltips and suggestions will need re-writing to JL standards anyway
        // show hover after pressing the modifier key
        wrapper.addEventListener('keydown', (event) => {
            if ((!hover_modifier || utils_1.getModifierState(event, hover_modifier)) &&
                this.hover_character === this.last_hover_character) {
                this.show_next_tooltip = true;
                this.handleHover(this.last_hover_response);
            }
        });
        this.editor.off('change', listeners.changeListener);
        // due to an unknown reason the default listener (as defined in the base class) is not invoked on file editors
        // the workaround - setting it at doc instead - works, thus the original one is first disabled (above) and a new
        // one is added (below); this however can have devastating effect on the editor synchronization - be careful!
        exports.CodeMirror.on(this.editor.getDoc(), 'change', (doc, change) => {
            this.handleChange(this.editor, change);
        });
    }
    get completionCharacters() {
        if (typeof this._completionCharacters === 'undefined' ||
            !this._completionCharacters.length) {
            this._completionCharacters = this.connection.getLanguageCompletionCharacters();
        }
        return this._completionCharacters;
    }
    get signatureCharacters() {
        if (typeof this._signatureCharacters === 'undefined' ||
            !this._signatureCharacters.length) {
            this._signatureCharacters = this.connection.getLanguageSignatureCharacters();
        }
        return this._signatureCharacters;
    }
    handleGoTo(locations) {
        this.remove_tooltip();
        // do NOT handle GoTo actions here
    }
    handleCompletion(completions) {
        // do NOT handle completion here
    }
    static get_markup_for_hover(response) {
        let contents = response.contents;
        // this causes the webpack to fail "Module not found: Error: Can't resolve 'net'" for some reason
        // if (lsProtocol.MarkedString.is(contents))
        ///  contents = [contents];
        if (typeof contents === 'string') {
            contents = [contents];
        }
        if (!Array.isArray(contents)) {
            return contents;
        }
        // now we have MarkedString
        let content = contents[0];
        if (typeof content === 'string') {
            // coerce to MarkedString  object
            return {
                kind: 'plaintext',
                value: content
            };
        }
        else {
            return {
                kind: 'markdown',
                value: '```' + content.language + '\n' + content.value + '```'
            };
        }
    }
    remove_range_highlight() {
        // @ts-ignore
        this._removeHover(); // this removes underlines
        this.last_hover_character = null;
    }
    remove_tooltip() {
        this.remove_range_highlight();
        if (this._tooltip !== undefined) {
            this._tooltip.dispose();
        }
    }
    handleHover(response) {
        this.remove_tooltip();
        this.last_hover_character = null;
        this.last_hover_response = null;
        if (!this.hover_character ||
            !response ||
            !response.contents ||
            (Array.isArray(response.contents) && response.contents.length === 0)) {
            return;
        }
        this.highlight_range(response.range, 'cm-lp-hover-available');
        if (!this.show_next_tooltip) {
            this.last_hover_response = response;
            this.last_hover_character = this.hover_character;
            return;
        }
        const markup = CodeMirrorAdapterExtension.get_markup_for_hover(response);
        let root_position = this.hover_character;
        let cm_editor = this.get_cm_editor(root_position);
        let editor_position = this.editor.root_position_to_editor_position(root_position);
        this._tooltip = this.create_tooltip(markup, cm_editor, editor_position);
    }
    get_cm_editor(position) {
        return this.editor.get_cm_editor(position);
    }
    get_language_at(position, editor) {
        return editor.getModeAt(position).name;
    }
    get_markup_for_signature_help(response, language) {
        let signatures = new Array();
        response.signatures.forEach((item) => {
            let markdown = '```' + language + '\n' + item.label + '\n```';
            if (item.documentation) {
                markdown += '\n';
                // TODO: make use of the MarkupContent object instead
                for (let line of item.documentation.toString().split('\n')) {
                    if (line.trim() === item.label.trim()) {
                        continue;
                    }
                    if (line.startsWith('>>>')) {
                        line = '```' + language + '\n' + line.substr(3) + '\n```';
                    }
                    markdown += line + '\n';
                }
            }
            signatures.push(markdown);
        });
        return {
            kind: 'markdown',
            value: signatures.join('\n\n')
        };
    }
    handleSignature(response) {
        this.remove_tooltip();
        if (!this.signature_character || !response || !response.signatures.length) {
            return;
        }
        let root_position = this.signature_character;
        let cm_editor = this.get_cm_editor(root_position);
        let editor_position = this.editor.root_position_to_editor_position(root_position);
        let language = this.get_language_at(editor_position, cm_editor);
        let markup = this.get_markup_for_signature_help(response, language);
        this._tooltip = this.create_tooltip(markup, cm_editor, editor_position);
    }
    updateAfterChange() {
        return __awaiter(this, void 0, void 0, function* () {
            this.remove_tooltip();
            yield utils_1.until_ready(() => this.last_change != null, 20, 25).catch(() => {
                this.invalidateLastChange();
                throw Error('No change obtained from CodeMirror editor within the expected time of 0.5s');
            });
            let change = this.last_change;
            try {
                const root_position = this.editor
                    .getDoc()
                    .getCursor('end');
                let document = this.editor.document_at_root_position(root_position);
                if (this.virtual_document !== document) {
                    return true;
                }
                if (!change || !change.text.length || !change.text[0].length) {
                    // deletion - ignore
                    return true;
                }
                let last_character;
                if (change.origin === 'paste') {
                    last_character = change.text[0][change.text.length - 1];
                }
                else {
                    last_character = change.text[0][0];
                }
                if (this.completionCharacters.indexOf(last_character) > -1) {
                    // TODO: pass info that we start from autocompletion (to avoid having . completion in comments etc)
                    //  it seems that it has to be done with a flag on a global object :(
                    this.invoke_completer();
                }
                else if (this.signatureCharacters.indexOf(last_character) > -1) {
                    this.signature_character = root_position;
                    let virtual_position = this.editor.root_position_to_virtual_position(root_position);
                    this.connection.getSignatureHelp(virtual_position);
                }
                return true;
            }
            catch (e) {
                console.log('handleChange failure - silent as to prevent editor going out of sync');
                console.error(e);
            }
            this.invalidateLastChange();
        });
    }
    invalidateLastChange() {
        this.last_change = null;
    }
    handleChange(cm, change) {
        this.last_change = change;
    }
    highlight_range(range, class_name) {
        let hover_character = this.hover_character;
        let start;
        let end;
        let start_in_editor;
        let end_in_editor;
        let cm_editor;
        // NOTE: foreign document ranges are checked before the request is sent,
        // no need to to this again here.
        if (range) {
            start = converter_1.PositionConverter.lsp_to_cm(range.start);
            end = converter_1.PositionConverter.lsp_to_cm(range.end);
            start_in_editor = this.virtual_document.transform_virtual_to_editor(start);
            end_in_editor = this.virtual_document.transform_virtual_to_editor(end);
            cm_editor = this.editor.get_editor_at_root_position(hover_character);
        }
        else {
            // construct range manually using the token information
            cm_editor = this.virtual_document.root.get_editor_at_source_line(hover_character);
            let token = this.editor.getTokenAt(hover_character);
            let start_in_root = {
                line: hover_character.line,
                ch: token.start
            };
            let end_in_root = {
                line: hover_character.line,
                ch: token.end
            };
            start_in_editor = this.editor.root_position_to_editor_position(start_in_root);
            end_in_editor = this.editor.root_position_to_editor_position(end_in_root);
        }
        // @ts-ignore
        this.hoverMarker = cm_editor
            .getDoc()
            .markText(start_in_editor, end_in_editor, { className: class_name });
    }
    position_from_mouse(ev) {
        return this.editor.coordsChar({
            left: ev.clientX,
            top: ev.clientY
        }, 'window');
    }
    is_token_empty(token) {
        return token.string.length === 0;
        // TODO  || token.type.length === 0? (sometimes the underline is shown on meaningless tokens)
    }
    // @ts-ignore
    _handleMouseOver(event) {
        // currently the events are coming from notebook panel; ideally these would be connected to individual cells,
        // (only cells with code) instead, but this is more complex to implement right. In any case filtering
        // is needed to determine in hovered character belongs to this virtual document
        let root_position = this.position_from_mouse(event);
        // happens because mousemove is attached to panel, not individual code cells,
        // and because some regions of the editor (between lines) have no characters
        if (typeof root_position === 'undefined') {
            this.remove_range_highlight();
            this.hover_character = null;
            return;
        }
        let token = this.editor.getTokenAt(root_position);
        let document = this.editor.document_at_root_position(root_position);
        let virtual_position = this.editor.root_position_to_virtual_position(root_position);
        if (this.is_token_empty(token) ||
            document !== this.virtual_document ||
            // @ts-ignore
            !this._isEventInsideVisible(event)) {
            this.remove_range_highlight();
            this.hover_character = null;
            return;
        }
        if (!positioning_1.is_equal(root_position, this.hover_character)) {
            this.hover_character = root_position;
            // @ts-ignore
            this.debouncedGetHover(virtual_position);
        }
    }
    handleMouseOver(event) {
        // proceed when no hover modifier or hover modifier pressed
        this.show_next_tooltip =
            !hover_modifier || utils_1.getModifierState(event, hover_modifier);
        try {
            return this._handleMouseOver(event);
        }
        catch (e) {
            if (!(e.message === 'Cell not found in cell_line_map' ||
                e.message === "Cannot read property 'string' of undefined")) {
                throw e;
            }
        }
    }
    collapse_overlapping_diagnostics(diagnostics) {
        // because Range is not a primitive types, the equality of the objects having
        // the same parameters won't be compared (thus considered equal) in Map.
        const range_id_to_range = new Map();
        const range_id_to_diagnostics = new Map();
        function get_range_id(range) {
            return (range.start.line +
                ',' +
                range.start.character +
                ',' +
                range.end.line +
                ',' +
                range.end.character);
        }
        diagnostics.forEach((diagnostic) => {
            let range = diagnostic.range;
            let range_id = get_range_id(range);
            range_id_to_range.set(range_id, range);
            if (range_id_to_diagnostics.has(range_id)) {
                let ranges_list = range_id_to_diagnostics.get(range_id);
                ranges_list.push(diagnostic);
            }
            else {
                range_id_to_diagnostics.set(range_id, [diagnostic]);
            }
        });
        let map = new Map();
        range_id_to_diagnostics.forEach((range_diagnostics, range_id) => {
            let range = range_id_to_range.get(range_id);
            map.set(range, range_diagnostics);
        });
        return map;
    }
    handleDiagnostic(response) {
        /* TODO: gutters */
        try {
            // Note: no deep equal for Sets or Maps in JS
            const markers_to_retain = new Set();
            // add new markers, keep track of the added ones
            // TODO: test for diagnostic messages not being over-writen
            //  test case: from statistics import mean, bisect_left
            //  and do not use either; expected: title has "mean imported but unused; bisect_left imported and unused'
            // TODO: test case for severity class always being set, even if diagnostic has no severity
            let diagnostics_by_range = this.collapse_overlapping_diagnostics(response.diagnostics);
            diagnostics_by_range.forEach((diagnostics, range) => {
                const start = converter_1.PositionConverter.lsp_to_cm(range.start);
                const end = converter_1.PositionConverter.lsp_to_cm(range.end);
                if (start.line > this.virtual_document.last_virtual_line) {
                    console.log('Malformed diagnostic was skipped (out of lines) ', diagnostics);
                    return;
                }
                // assuming that we got a response for this document
                let start_in_root = this.transform_virtual_position_to_root_position(start);
                let document = this.editor.document_at_root_position(start_in_root);
                // TODO why do I get signals from the other connection in the first place?
                //  A: because each virtual document adds listeners AND if the extracted content
                //  is kept in the host document, it remains...
                if (this.virtual_document !== document) {
                    console.log(`Ignoring inspections from ${response.uri}`, ` (this region is covered by a another virtual document: ${document.uri})`, ` inspections: `, diagnostics);
                    return;
                }
                if (document.virtual_lines
                    .get(start.line)
                    .skip_inspect.indexOf(document.id_path) !== -1) {
                    console.log('Ignoring inspections silenced for this document:', diagnostics);
                    return;
                }
                let highest_severity_code = diagnostics
                    .map(diagnostic => diagnostic.severity || default_severity)
                    .sort()[0];
                const severity = lsp_1.diagnosticSeverityNames[highest_severity_code];
                let cm_editor = document.get_editor_at_virtual_line(start);
                let start_in_editor = document.transform_virtual_to_editor(start);
                let end_in_editor = document.transform_virtual_to_editor(end);
                // what a pity there is no hash in the standard library...
                // we could use this: https://stackoverflow.com/a/7616484 though it may not be worth it:
                //   the stringified diagnostic objects are only about 100-200 JS characters anyway,
                //   depending on the message length; this could be reduced using some structure-aware
                //   stringifier; such a stringifier could also prevent the possibility of having a false
                //   negative due to a different ordering of keys
                // obviously, the hash would prevent recovery of info from the key.
                let diagnostic_hash = JSON.stringify({
                    // diagnostics without ranges
                    diagnostics: diagnostics.map(diagnostic => [
                        diagnostic.severity,
                        diagnostic.message,
                        diagnostic.code,
                        diagnostic.source,
                        diagnostic.relatedInformation
                    ]),
                    // the apparent marker position will change in the notebook with every line change for each marker
                    // after the (inserted/removed) line - but such markers should not be invalidated,
                    // i.e. the invalidation should be performed in the cell space, not in the notebook coordinate space,
                    // thus we transform the coordinates and keep the cell id in the hash
                    range: {
                        start: start_in_editor,
                        end: end_in_editor
                    },
                    editor: this.unique_editor_ids.get(cm_editor)
                });
                markers_to_retain.add(diagnostic_hash);
                if (!this.marked_diagnostics.has(diagnostic_hash)) {
                    let options = {
                        title: diagnostics
                            .map(d => d.message + (d.source ? ' (' + d.source + ')' : ''))
                            .join('\n'),
                        className: 'cm-lsp-diagnostic cm-lsp-diagnostic-' + severity
                    };
                    let marker;
                    try {
                        marker = cm_editor
                            .getDoc()
                            .markText(start_in_editor, end_in_editor, options);
                    }
                    catch (e) {
                        console.warn('Marking inspection (diagnostic text) failed, see following logs (2):');
                        console.log(diagnostics);
                        console.log(e);
                        return;
                    }
                    this.marked_diagnostics.set(diagnostic_hash, marker);
                }
            });
            // remove the markers which were not included in the new message
            this.marked_diagnostics.forEach((marker, diagnostic_hash) => {
                if (!markers_to_retain.has(diagnostic_hash)) {
                    this.marked_diagnostics.delete(diagnostic_hash);
                    marker.clear();
                }
            });
        }
        catch (e) {
            console.warn(e);
        }
    }
    transform_virtual_position_to_root_position(start) {
        let cm_editor = this.virtual_document.virtual_lines.get(start.line).editor;
        let editor_position = this.virtual_document.transform_virtual_to_editor(start);
        return this.editor.transform_editor_to_root(cm_editor, editor_position);
    }
}
exports.CodeMirrorAdapterExtension = CodeMirrorAdapterExtension;
